// A slightly improved crafting tool.  You can click on
// the faces of the cubes to build outward from those
// faces.  Uses the sneaky trick from Module02 for
// direct manipulation: draw a false colour image in 
// an offscreen buffer and look up pixels in that buffer
// to decide where the user clicked.  In this case, I
// sneakily code the pixels of the false colour image
// to be the integer coordinates where new cubes ought
// to be added.  This is really, really devious, and I'm
// tempted not to share it, but here it is.  Just 
// understand that this is well beyond the level of 
// complexity I'd expect students in this course to achieve
// themselves.

class Point
{
  int x;
  int y;
  int z;

  Point( int x, int y, int z )
  {
    this.x = x;
    this.y = y;
    this.z = z;
  }

  // Ask a cube to draw itself in two different ways.
  // In the simple version, go ahead and draw the cube
  // in the normal way.  In the complicated version,
  // draw each of the faces of the cube separately, so
  // that they can all have their own colours (for picking
  // purposes).  
  void draw( PGraphics g, boolean pick )
  {
    g.pushMatrix();
    g.translate( x, y, z );

    if ( pick ) {
      // Back
      g.pushMatrix();
      g.translate( 0, 0, -0.5 );
      g.fill( x + 128, y + 128, z + 128 - 1 );
      g.rect( -0.5, -0.5, 1, 1 );
      g.popMatrix();

      // Front
      g.pushMatrix();
      g.rotateY( PI );
      g.translate( 0, 0, -0.5 );
      g.fill( x + 128, y + 128, z + 128 + 1 );
      g.rect( -0.5, -0.5, 1, 1 );
      g.popMatrix();
      
      g.pushMatrix();
      g.rotateY( HALF_PI );
      
      // Left
      g.pushMatrix();
      g.translate( 0, 0, -0.5 );
      g.fill( x + 128 - 1, y + 128, z + 128 );
      g.rect( -0.5, -0.5, 1, 1 );
      g.popMatrix();

      // Right
      g.pushMatrix();
      g.rotateY( PI );
      g.translate( 0, 0, -0.5 );
      g.fill( x + 128 + 1, y + 128, z + 128 );
      g.rect( -0.5, -0.5, 1, 1 );
      g.popMatrix();
      g.popMatrix();

      g.pushMatrix();
      g.rotateX( HALF_PI );
      
      // Top
      g.pushMatrix();
      g.translate( 0, 0, -0.5 );
      g.fill( x + 128, y + 128 + 1, z + 128 );
      g.rect( -0.5, -0.5, 1, 1 );
      g.popMatrix();

      // Bottom
      g.pushMatrix();
      g.rotateY( PI );
      g.translate( 0, 0, -0.5 );
      g.fill( x + 128, y + 128 - 1, z + 128 );
      g.rect( -0.5, -0.5, 1, 1 );
      g.popMatrix();
      g.popMatrix();
    } else {
      box( 1 );
    }

    g.popMatrix();
  }
}

Point[] cubes;
Point cur;
PGraphics offs;

boolean dragging = false;
float rx = 0.0;
float ry = 0.0;

void init()
{
  cubes = new Point[1];
  cubes[0] = new Point( 0, 0, 0 );
  cur = new Point( 0, 0, 0 );
}

void setup()
{
  size( 500, 500, P3D );
  init();
  
  // Create an offscreen buffer
  offs = createGraphics( 500, 500, P3D );
}

void draw()
{
  background( 50 );
  lights();
  fill( 200, 200, 240 );
  noStroke();

  translate( width/2, height/2 );
  // Use crappy X and Y rotation.  Sorry.
  rotateY( ry );
  rotateX( rx );
  scale( 50 );

  for ( Point p : cubes ) {
    p.draw( g, false );
  }
}

// When the user clicks, draw to the offscreen buffer
// and decide if any cube face was clicked.  If so,
// add a new cube.  If not, assume the user wants to
// rotate.
void mousePressed()
{
  offs.noSmooth();

  offs.beginDraw();
  offs.background( 255 );
  offs.noLights();
  offs.noStroke();

  offs.translate( width/2, height/2 );
  offs.rotateY( ry );
  offs.rotateX( rx );
  offs.scale( 50 );

  for ( Point p : cubes ) {
    p.draw( offs, true );
  }
  offs.endDraw();

  // Read the pixel at the mouse position...
  color c = offs.get( mouseX, mouseY );
  // And decode its colour into magic X, Y and Z
  // coordinates.
  int nx = int( red( c ) - 128 );
  int ny = int( green( c ) - 128 );
  int nz = int( blue( c ) - 128 );
  //println( nx, ny, nz );

  if ( abs( nx ) < 100 && abs( ny ) < 100 && abs( nz ) < 100 ) {
    // Add a cube.
    cubes = (Point[])append( cubes, new Point( nx, ny, nz ) );
  } else {
    dragging = true;
  }
}

void mouseDragged()
{
  if( dragging ) {
    ry += (mouseX - pmouseX) * 0.01;
    rx += (mouseY - pmouseY) * 0.01;
  }
}

void mouseReleased()
{
  dragging = false;
}